; vim:syntax=avrm48:
; ds3231-lcd.asm  DS3231 auslesen und auf lcd ausgeben
; interner Osci 8 MHz
; laufend UTC Datum formatiert ausgeben
; automatischer Wechsel Sommerzeit - Winterzeit folgt
; Beginn: 23.4.2015 Stand: 26.10.2015
; Author: Alfred Hell alhell@ajhe.eu

.include "/usr/share/avra/m48def.inc"

.def	dat1	= r9
.def	dat2	= r10
.def	dat3	= r11
.def	dat4	= r12
.def	dat5	= r13
.def	dat6	= r14
.def	temp  	= r16
.def	zaeh1	= r17
.def	rmp	= r18
.def	aus1	= r19
.def	aus2	= r20
.def	temp1	= r21
.def	temp2	= r22
.def	temp3	= r23

.equ	somwin	= 1		; 1 = UTC + 1, 2 = UTC +2

.dseg
zeittemp:	.byte	18
wochdat:	.byte	18
einles:		.byte	20

.cseg
.org 0x00
	rjmp	start

.org  	INT_VECTORS_SIZE		; Platz fuer unbenutzte Interupts	

start:
	ldi	rmp, HIGH(RAMEND)
	out	SPH, rmp
	ldi	rmp, LOW(RAMEND)	; Stack initialisieren
	out	SPL, rmp
	ldi	rmp, $ff
	out	DDRD, rmp		; Port D Ausgang

	rcall	lcd_init

	ldi	rmp, 32			; Bitrate 100 kHz
	sts	TWBR, rmp
anfang1:
	ldi	XL, low(einles)
	ldi	XH, high(einles)
	
	ldi	rmp, (1<<TWINT)|(1<<TWSTA)|(1<<TWEN)
	sts	TWCR, rmp		; einschalten und Startkommando senden
	rcall	TWwart			; warten bis fertig
	lds	rmp, TWSR		; testen ob OK
	andi	rmp, 0xF8		; maskieren lt. Datenblatt
	cpi	rmp, 0x08		; wenn nicht dann Fehler
	breq	wei1	
	rcall	TWerr
	rjmp	anfang1			; nach Fehler von vorne
wei1:	
	ldi	rmp, 0b11010000		; Bit 0 = 1 = empfangen, Bit 7-4 = Adresse slave
	sts	TWDR, rmp		; Bit 1 - 3 null Adresse
	ldi	rmp, (1<<TWINT)|(1<<TWEN)
	sts	TWCR, rmp		; warten auf Antwort
	rcall	TWwart
	lds	rmp, TWSR		; testen
	andi	rmp, 0xF8
	cpi	rmp, 0x18		; Antwort OK, dann weiter
	breq	wei2	
	rcall	TWerr
	rjmp	anfang1			; nach Fehler von vorne
wei2:	
	clr	rmp			; zuerst Adresse auf 0 setzen
	sts	TWDR, rmp		; Adresse bit 0 - 7
	ldi	rmp, (1<<TWINT)|(1<<TWEN)
	sts	TWCR, rmp		; warten auf Antwort
	rcall	TWwart
	lds	rmp, TWSR		; testen
	andi	rmp, 0xF8
	cpi	rmp, 0x28		; Antwort OK, dann weiter
	breq	wei3	
	rcall	TWerr
	rjmp	anfang1			; nach Fehler von vorne
wei3:
	ldi	rmp, (1<<TWINT)|(1<<TWSTO)|(1<<TWEN)
	sts	TWCR, rmp		; TW abschalten

anfang:
	ldi	XL, low(einles)
	ldi	XH, high(einles)
	ldi	rmp, (1<<TWINT)|(1<<TWSTA)|(1<<TWEN)
	sts	TWCR, rmp		; einschalten und Startkommando senden
	rcall	TWwart			; warten bis fertig
	lds	rmp, TWSR		; testen ob OK
	andi	rmp, 0xF8		; maskieren lt. Datenblatt
	cpi	rmp, 0x08		; wenn nicht dann Fehler
	breq	wei4	
	rcall	TWerr
	rjmp	anfang			; nach Fehler von vorne
wei4:	
	ldi	rmp, 0b11010001		; Bit 0 = 1 = empfangen, Bit 7-4 = Adresse slave
	sts	TWDR, rmp		; Bit 1 - 3 null Adresse
	ldi	rmp, (1<<TWINT)|(1<<TWEN)
	sts	TWCR, rmp		; warten auf Antwort
	rcall	TWwart
	lds	rmp, TWSR		; testen
	andi	rmp, 0xF8
	cpi	rmp, 0x40		; Antwort OK, dann weiter
	breq	wei5	
	rcall	TWerr
	rjmp	anfang			; nach Fehler von vorne
wei5:	
	ldi	zaeh1, 18
noch:
	ldi	rmp, (1<<TWINT)|(1<<TWEA)|(1<<TWEN)
	sts	TWCR, rmp		; ack senden
	rcall	TWwart
	lds	rmp, TWSR		; testen
	andi	rmp, 0xF8
	cpi	rmp, 0x50		; Daten empfangen OK?
	breq	wei6	
	rcall	TWerr
	rjmp	anfang			; nach Fehler von vorne
wei6:	
	lds	temp, TWDR		; einlesen
	st	X+, temp
	dec	zaeh1
	tst	zaeh1			; 
	brne	noch			; Zeit und Datum auslesen 

	ldi	rmp, (1<<TWINT)|(0<<TWEA)|(1<<TWEN)
	sts	TWCR, rmp		; kein ack senden
	rcall	TWwart			; dadurch wird auslesen beendet
	lds	rmp, TWSR
	andi	rmp, 0xF8
	cpi	rmp, 0x58		; letztes byte empfangen OK?
	breq	wei7	
	rcall	TWerr
	rjmp	anfang			; nach Fehler von vorne
wei7:	
	lds	temp, TWDR
	st	X+, temp
	ldi	rmp, (1<<TWINT)|(1<<TWSTO)|(1<<TWEN)
	sts	TWCR, rmp		; TW abschalten
	rcall	lcd_clear
	rcall 	ausgeb
	rcall	warte
	rjmp    anfang			; nochmal von vorne
	
TWwart:	
	lds	rmp, TWCR
	sbrs	rmp, TWINT		; warten bist TWINT Bit gesetzt
	rjmp	TWwart
	ret

TWerr:
	ldi	rmp, (1<<TWINT)|(1<<TWSTO)|(1<<TWEN)
	sts	TWCR, rmp		; Stop senden
	rcall	TWwart
	clr	rmp
	sts	TWCR, rmp		; und TWI abschalten
	ret

warte:
	push	temp1
	push	temp2
	push	temp3
	ldi	temp3, 90		; = ca. 2 Sekunden
	clr	temp2
	clr	temp1
warte1:
	dec	temp1
	brne	warte1
	dec	temp2
	brne	warte1
	dec	temp3
	brne	warte1
	pop	temp3
	pop	temp2
	pop 	temp1
	ret

ausgeb:
	ldi	XL, low(einles)		; einlesebuffer
	ldi	XH, high(einles)
	ldi	YL, low(zeittemp)	; Buffer fuer Formatierung
	ldi	YH, high(zeittemp)
	ldi	ZL, low(wochdat)	; Buffer fuer Formatierung
	ldi	ZH, high(wochdat)
	ld	temp, X+		; lade 1. Byte = Sekunden
	rcall	wandle			; BCD fuer Ausgabe umwandeln
	mov	dat5, aus1		; in Register fuer umreihung
	mov	dat6, aus2
	ld	temp, X+		; Minuten
	rcall	wandle
	mov	dat3, aus1
	mov	dat4, aus2
	ld	temp, X+		; Stunden
	rcall	sowi			; add Sommer oder Winterzeit
	rcall	wandle
	mov	dat1, aus1
	mov	dat2, aus2
	st	Y+, dat1		; formatiert in Buffer
	st	Y+, dat2		; Stunden
	ldi	rmp, 0x3a		; = :
	st	Y+, rmp
	st	Y+, dat3		; Minuten
	st	Y+, dat4
	ldi	rmp, 0x3a
	st	Y+, rmp
	st	Y+, dat5		; Sekunden
	st	Y+, dat6
	ldi	rmp, 0x20		; Leerzeichen
	st	Y+, rmp
	ld	temp, X+		; lade Wochentag
	rcall	wandlewota
	st	Z+, aus1		; wandeln und in Buffer
	st	Z+, aus2		; wandeln und in Buffer
	ldi	rmp, 0x2e		; Punkt
	st	Z+, rmp
	ldi	rmp, 0x20
	st	Z+, rmp
	st	Z+, rmp			; Leerzeichen
	st	Z+, rmp
	ld	temp, X+		; lade Tag
	rcall	wandle
	mov	dat1, aus1
	mov	dat2, aus2
	ld	temp, X+		; Monat
	andi	temp, 0x7f		; century bit loeschen
	rcall	wandle
	mov	dat3, aus1
	mov	dat4, aus2
	ld	temp, X+		; Jahr
	rcall	wandle
	mov	dat5, aus1
	mov	dat6, aus2
	st	Z+, dat1		; in Buffer schreiben
	st	Z+, dat2		; Tag
	ldi	rmp, 0x2e
	st	Z+, rmp
	st	Z+, dat3		; Monat
	st	Z+, dat4
	ldi	rmp, 0x2e
	st	Z+, rmp
	ldi	rmp, 0x32		; Jahr 2
	st	Z+, rmp
	ldi	rmp, 0x30		; Jahr 0
	st	Z+, rmp
	st	Z+, dat5		; Jahr 15
	st	Z+, dat6
	adiw	XH:XL, 10
	ld	temp, X+		; lade Temp signed!
	sbrc	temp, 7			; Vorzeichen testen
	rjmp	naech1
	ldi	rmp, 0x2b		; Plus
	st	Y+, rmp
	rjmp	naech2
naech1:	neg	temp
	ldi	rmp, 0x2d		; Minus
	st	Y+, rmp
naech2:	ldi	rmp, 0x2F		; Temp umwandeln
numb1:	inc	rmp			; und in Buffer
	subi	temp, 100
	brcc	numb1
	subi	temp, -100
;	st	Y+, rmp
	ldi	rmp, 0x2F
numb2:	inc	rmp
	subi	temp, 10
	brcc	numb2
	subi	temp, -10
	st	Y+, rmp
	ldi	rmp, 0x30
	add	rmp, temp
	st	Y+, rmp
	ldi	rmp, 0x2e		; Punkt
	st	Y+, rmp
	ld	temp, X			; nachkommastellen laden
	cpi	temp, 0x40		; und umwandeln
	brne	spri1
	ldi	rmp, 0x32		; x,25
	st	Y+, rmp
	ldi	rmp, 0x35
	st	Y+, rmp
	rjmp 	linfee
spri1:	cpi	temp, 0x80
	brne	spri2
	ldi	rmp, 0x35		; x,50
	st	Y+, rmp
	ldi	rmp, 0x30
	st	Y+, rmp
	rjmp	linfee
spri2:	cpi	temp, 0xC0
	brne	spri3
	ldi	rmp, 0x37		; x,75
	st	Y+, rmp
	ldi	rmp, 0x35
	st	Y+, rmp
	rjmp	linfee
spri3:	ldi	rmp, 0x30
	st	Y+, rmp
	st	Y+, rmp
linfee:	ldi	rmp, 0x43		; = C
	st	Y+, rmp

	rcall	lcd_home
	ldi	YL, low(zeittemp)	; Buffer fuer Formatierung
	ldi	YH, high(zeittemp)
	ldi	zaeh1, 16
noma:	ld	temp1, Y+
	rcall	lcd_data		; Daten ausgeben
	dec	zaeh1
	tst	zaeh1	
	brne	noma

	rcall	lcd_zeile2
	ldi	ZL, low(wochdat)	; Buffer fuer Formatierung
	ldi	ZH, high(wochdat)
	ldi	zaeh1, 16
noma1:	ld	temp1, Z+
	rcall	lcd_data		; Daten ausgeben
	dec	zaeh1
	tst	zaeh1
	brne	noma1
	ret
	
wandle:	mov	temp1, temp
	mov	temp2, temp
	ldi	temp3, 0x30		; umwandeln	
	swap	temp1
	andi	temp1, 0x0f		; obere loeschen
	add	temp1, temp3
	mov	aus1, temp1
	andi	temp2, 0x0f		; ober loeschen
	add	temp2, temp3		; umwandeln
	mov	aus2, temp2
	ret

sowi:	push	temp1
	ldi	temp1, somwin		; add diff (somwin)
	add	temp, temp1		; add MEZ or MESZ
	cpi 	temp, 0x0a		; => 10
	brcs	sw1			; < nichts tun
	breq	gleich			; ist gleich, korrigieren 
	cpi	temp, 0x0b		; == 11
	brne	noelf
	ldi	temp, 0x11		; BCD wandeln
noelf:	cpi	temp, 0x1a		; => 20
	brcs	sw1			; < nichts tun	
	breq	glei			; ist gleich, korrigieren
	cpi	temp, 0x1b		; == 21
	brne	unzeh
	ldi	temp, 0x21		; BCD wandeln
	rjmp	sw1
glei:	ldi	temp, 0x20		; BCD wandeln
	rjmp	sw1
gleich:	ldi	temp, 0x10		; BCD wandeln
unzeh:	sbrs	temp, 5			; unter 20 Uhr springe
	rjmp	sw1
	cpi	temp, 0x24		; Vergleich mit 24
	brne	sw2			; 
	clr	temp			; wenn 24 dann 0 Uhr
sw2:	cpi	temp, 0x25		; Vergleich mit 25
	brne	sw1
	ldi	temp, 1			; wenn 25 dann 1 Uhr
sw1:	pop	temp1
	ret

wandlewota:
	cpi	temp, 2
	brne	dita
	ldi	rmp, 0x4d		; Mo = Montag
	mov	aus1, rmp
	ldi	rmp, 0x6f
	mov	aus2, rmp
	ret
dita:	cpi	temp, 3
	brne	mita
	ldi	rmp, 0x44		; Di = Dienstag
	mov	aus1, rmp
	ldi	rmp, 0x69
	mov	aus2, rmp
	ret
mita:	cpi	temp, 4
	brne	dota
	ldi	rmp, 0x4d		; Mi = Mittwoch
	mov	aus1, rmp
	ldi	rmp, 0x69
	mov	aus2, rmp
	ret
dota:	cpi	temp, 5
	brne	frta
	ldi	rmp, 0x44		; Do = Donnerstag
	mov	aus1, rmp
	ldi	rmp, 0x6f
	mov	aus2, rmp
	ret
frta:	cpi	temp, 6
	brne	sata
	ldi	rmp, 0x46		; Fr = Freitag
	mov	aus1, rmp
	ldi	rmp, 0x72
	mov	aus2, rmp
	ret
sata:	cpi	temp, 7
	brne	sota
	ldi	rmp, 0x53		; Sa = Samstag
	mov	aus1, rmp
	ldi	rmp, 0x61
	mov	aus2, rmp
	ret
sota:	cpi	temp, 1
	brne	nixta
	ldi	rmp, 0x53		; So = Sonntag
	mov	aus1, rmp
	ldi	rmp, 0x6f
	mov	aus2, rmp
	ret
nixta:  ldi	rmp, 0x20
	mov	aus1, rmp
	mov	aus2, rmp
	ret

;sendet ein Datenbyte an das LCD
lcd_data:
           mov temp2, temp1             ; "Sicherungskopie" für
                                        ; die Übertragung des 2.Nibbles
           swap temp1                   ; Vertauschen
           andi temp1, 0b00001111       ; oberes Nibble auf Null setzen
           sbr temp1, 1<<4              ; entspricht 0b00010000 (Anm.1)
           out PORTD, temp1             ; ausgeben
           rcall lcd_enable             ; Enable-Routine aufrufen
                                        ; 2. Nibble, kein swap da es schon
                                        ; an der richtigen stelle ist
           andi temp2, 0b00001111       ; obere Hälfte auf Null setzen 
           sbr temp2, 1<<4              ; entspricht 0b00010000
           out PORTD, temp2             ; ausgeben
           rcall lcd_enable             ; Enable-Routine aufrufen
           rcall delay50us              ; Delay-Routine aufrufen
           ret                          ; zurück zum Hauptprogramm
 
; sendet einen Befehl an das LCD
lcd_command:                            ; wie lcd_data, nur RS=0
           mov temp2, temp1
           swap temp1
           andi temp1, 0b00001111
           out PORTD, temp1
           rcall lcd_enable
           andi temp2, 0b00001111
           out PORTD, temp2
           rcall lcd_enable
           rcall delay50us
           ret
 
; erzeugt den Enable-Puls
lcd_enable:
	   nop
	   nop
           sbi PORTD, 5                 ; Enable high
           nop                          ; mindestens 3 Taktzyklen warten
           nop
	   nop
	   nop
           nop
           cbi PORTD, 5                 ; Enable wieder low
           ret                          ; Und wieder zurück                     

; Pause nach jeder Übertragung
delay50us:                              ; 50us Pause (bei 8 MHz)
           ldi  temp1, $85
delay50us_:dec  temp1
           brne delay50us_
           ret                          ; wieder zurück
 
; Längere Pause für manche Befehle
delay5ms:                               ; 5ms Pause (bei 8 MHz)
           ldi  temp1, $67		; 21 bei 4MHz
WGLOOP0:   ldi  temp2, $FE		; C9 bei 4MHz
WGLOOP1:   dec  temp2
           brne WGLOOP1
           dec  temp1
           brne WGLOOP0
           ret                          ; wieder zurück
 
; Initialisierung: muss ganz am Anfang des Programms aufgerufen werden
lcd_init:
           ldi  temp3,50
powerupwait:
           rcall  delay5ms
           dec  temp3
           brne powerupwait
           ldi temp1, 0b00000011        ; muss 3mal hintereinander gesendet
           out PORTD, temp1             ; werden zur Initialisierung
           rcall lcd_enable             ; 1
           rcall delay5ms
           rcall lcd_enable             ; 2
           rcall delay5ms
           rcall lcd_enable             ; und 3!
           rcall delay5ms
           ldi temp1, 0b00000010        ; 4bit-Modus einstellen
           out PORTD, temp1
           rcall lcd_enable
           rcall delay5ms
           ldi temp1, 0b00101000        ; 4Bit / 2 Zeilen / 5x8
           rcall lcd_command
           ldi temp1, 0b00001100        ; Display ein / Cursor aus / kein Blinken
           rcall lcd_command
           ldi temp1, 0b00000100        ; inkrement / kein Scrollen
           rcall lcd_command
           ret
 
; Sendet den Befehl zur Löschung des Displays
lcd_clear:
           ldi temp1, 0b00000001   ; Display löschen
           rcall lcd_command
           rcall delay5ms
           ret
 
; Sendet den Befehl: Cursor Home
lcd_home:
           ldi temp1, 0b00000010   ; Cursor Home
           rcall lcd_command
           rcall delay5ms
           ret

; cursor 2. Zeile
lcd_zeile2:
           push  temp1
           ldi   temp1,    0b11000000   ; Cursor Zeil2
           rcall lcd_command
           rcall delay5ms
           pop   temp1
           ret
